{
(***********************************************************************)
(*                                                                     *)
(*                           The V6 Engine                             *)
(*                                                                     *)
(*          Francois Rouaix, projet Cristal, INRIA Rocquencourt        *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: lexdate.mll,v 1.2 1998/08/25 16:27:25 rouaix Exp $ *)

open Mstring
open Errors
open Http_date
}

rule wkday = parse
   "Sun" {0}
 | "Mon" {1}
 | "Tue" {2}
 | "Wed" {3}
 | "Thu" {4}
 | "Fri" {5}
 | "Sat" {6}
 | "" {error (Lexical ("invalid wkday", Lexing.lexeme_start lexbuf))}

and weekday = parse
   "Sunday" {0}
 | "Monday" {1}
 | "Tuesday" {2}
 | "Wednesday" {3}
 | "Thursday" {4}
 | "Friday" {5}
 | "Saturday" {6}
 | "" {error (Lexical ("invalid weekday", Lexing.lexeme_start lexbuf))}

and month = parse
   "Jan" {0}
 | "Feb" {1}
 | "Mar" {2}
 | "Apr" {3}
 | "May" {4}
 | "Jun" {5}
 | "Jul" {6}
 | "Aug" {7}
 | "Sep" {8}
 | "Oct" {9}
 | "Nov" {10}
 | "Dec" {11}
 | "" {error (Lexical ("invalid month", Lexing.lexeme_start lexbuf))}

and twodigit = parse
  ['0'-'9']['0'-'9'] { int_of_string (Lexing.lexeme lexbuf)}
 | "" {error (Lexical ("invalid 2DIGIT", Lexing.lexeme_start lexbuf))}

and fourdigit = parse
  ['0'-'9']['0'-'9']['0'-'9']['0'-'9'] 
      { int_of_string (Lexing.lexeme lexbuf)}
 | "" {error (Lexical ("invalid 4DIGIT", Lexing.lexeme_start lexbuf))}

(* longest match *)
and twoorfourdigit = parse
  ['0'-'9']['0'-'9']['0'-'9']['0'-'9'] 
      { int_of_string (Lexing.lexeme lexbuf), true}
 | ['0'-'9']['0'-'9'] 
      { int_of_string (Lexing.lexeme lexbuf), false}
 | "" {error (Lexical ("invalid 2/4DIGIT", Lexing.lexeme_start lexbuf))}

and wdigit = parse
  ' ' ['0'-'9'] 
    { let s = Lexing.lexeme lexbuf in Char.code s.[1] - Char.code '0' }
 | ['0'-'9']['0'-'9'] 
    { int_of_string (Lexing.lexeme lexbuf)}
 | "" {error (Lexical ("invalid 2DIGIT", Lexing.lexeme_start lexbuf))}

and comma = parse
   ',' { () }
 | "" {error (Lexical ("comma expected", Lexing.lexeme_start lexbuf))}

and semicolon = parse
  ';' { () }
 | "" {error (Lexical ("semicolon expected", Lexing.lexeme_start lexbuf))}

and colon = parse
  ':' { () }
 | "" {error (Lexical ("semicolon expected", Lexing.lexeme_start lexbuf))}

and sp = parse
   ' ' { () }
 | "" {error (Lexical ("SP expected", Lexing.lexeme_start lexbuf))}

and hyphen = parse
   '-' { () }
 | "" {error (Lexical ("SP expected", Lexing.lexeme_start lexbuf))}

and gmt = parse
   "GMT" { () }
 | "" {error (Lexical ("GMT expected", Lexing.lexeme_start lexbuf))}

{

let parse_time lexbuf =
  let hour = twodigit lexbuf in
  let _ = colon lexbuf in
  let min = twodigit lexbuf in
  let _ = colon lexbuf in
  let sec = twodigit lexbuf in
    (hour, min, sec)

let lenient_year lexbuf =
  match twoorfourdigit lexbuf with
    n, true -> n - 1900
  | n, false -> n

(* Fri, 14 Sep 1998 01:02:03 GMT *)
let rfc1123 lexbuf =
  let wday = wkday lexbuf in
  let _ = comma lexbuf in
  let _ = sp lexbuf in
  (* date1 *)
  let mday = twodigit lexbuf in
  let _ = sp lexbuf in
  let mon = month lexbuf in
  let _ = sp lexbuf in
  let year = lenient_year lexbuf in
  let _ = sp lexbuf in
  let hour,min,sec = parse_time lexbuf in
  let _ = sp lexbuf in
  let _ = gmt lexbuf in
    {ht_sec = sec;
     ht_min = min;
     ht_hour = hour;
     ht_mday = mday;
     ht_mon = mon;
     ht_year = year;
     ht_wday = wday }

(* Friday, 14-Sep-98 01:02:03 GMT *)
let rfc850 lexbuf =
  let wday = weekday lexbuf in
  let _ = comma lexbuf in
  let _ = sp lexbuf in
  (* date2 *)
  let mday = twodigit lexbuf in
  let _ = hyphen lexbuf in
  let mon = month lexbuf in
  let _ = hyphen lexbuf in
  let year = lenient_year lexbuf in
  let _ = sp lexbuf in
  let hour,min,sec = parse_time lexbuf in
  let _ = sp lexbuf in
  let _ = gmt lexbuf in
    {ht_sec = sec;
     ht_min = min;
     ht_hour = hour;
     ht_mday = mday;
     ht_mon = mon;
     ht_year = year;
     ht_wday = wday }

(* Fri Aug 14 01:02:03 1998 *)
let asctime lexbuf =
  let wday = wkday lexbuf in
  let _ = sp lexbuf in
  (* date3 *)
  let mon = month lexbuf in
  let _ = sp lexbuf in
  let mday = wdigit lexbuf in
  let _ = sp lexbuf in
  let hour,min,sec = parse_time lexbuf in
  let _ = sp lexbuf in
  let year = lenient_year lexbuf in
    {ht_sec = sec;
     ht_min = min;
     ht_hour = hour;
     ht_mday = mday;
     ht_mon = mon;
     ht_year = year;
     ht_wday = wday }

(* RFC850 has -
 * RFC850 and 822 have commas
 * otherwise it's asctime
 *)

let ht_of_string s =
  try
    let _ = String.index s '-' in
      rfc850 (Lexing.from_string s)
  with
    Not_found ->
      try
        let _ = String.index s ',' in
          rfc1123 (Lexing.from_string s)
      with
      	Not_found -> 
          asctime (Lexing.from_string s)



}
