(********************************************************************** *)
(*                                                                      *)
(*                        Ubi Est - Indexing package                    *)
(*                                                                      *)
(*  Copyright Franois Pessaux 1999-2007.                               *)
(*  Distributed according to the LICENCE file contained in the archive. *)
(*                                                                      *)
(************************************************************************)


open Infinity ;;


type extent_position = int ;;
type extent = (extent_position * extent_position) ;;


exception Found of extent ;;



let sequential_search_tau index_data token k =
  if k = !plus_infty then (!plus_infty, !plus_infty)
  else
    try
      let positions = Hashtbl.find index_data token in
      let str_len = (Token.length token - 1) in
      for i = 0 to Array.length positions - 1 do
	let p = positions.(i) in
	(* Ne pas oublier de grer les infinis locaux *)
	if p >= !plus_infty then raise Not_found ;
	if k <= p && p > !minus_infty then
	  let q = p + str_len in raise (Found (p, q))
      done ;
      (!plus_infty, !plus_infty)
    with Found ext -> ext | Not_found -> (!plus_infty, !plus_infty)
 ;;



let sequential_search_rho index_data token k =
  if k = !plus_infty then (!plus_infty, !plus_infty)
  else
    try
      let positions = Hashtbl.find index_data token in
      let str_len = (Token.length token - 1) in
      for i = 0 to Array.length positions - 1 do
	let p = positions.(i) in
	let q = p + str_len in
	(* Ne pas oublier de grer les infinis locaux *)
	if q >= !plus_infty then raise Not_found ;
	if k <= q && q > !minus_infty then raise (Found (p, q))
      done ;
      (!plus_infty, !plus_infty)
    with Found ext -> ext | Not_found -> (!plus_infty, !plus_infty)
 ;;



let sequential_search_tau' index_data token k =
  if k = !minus_infty then (!minus_infty, !minus_infty)
  else
    try
      let positions = Hashtbl.find index_data token in
      let str_len = (Token.length token - 1) in
      for i = Array.length positions - 1 downto 0 do
	let p = positions.(i) in
	let q = p + str_len in
	(* Ne pas oublier de grer les infinis locaux *)
	if q <= !minus_infty then raise Not_found ;
	if k >= q && q < !plus_infty then raise (Found (p, q))
      done ;
      (!minus_infty, !minus_infty)
    with Found ext -> ext | Not_found -> (!minus_infty, !minus_infty)
 ;;



let sequential_search_rho' index_data token k =
  if k = !minus_infty then (!minus_infty, !minus_infty)
  else
    try
      let positions = Hashtbl.find index_data token in
      let str_len = (Token.length token - 1) in
      for i = Array.length positions - 1 downto 0 do
	let p = positions.(i) in
	(* Ne pas oublier de grer les infinis locaux *)
	if p <= !minus_infty then raise Not_found ;
	if k >= p && p < !plus_infty then
	  let q = p + str_len in raise (Found (p, q))
      done ;
      (!minus_infty, !minus_infty)
    with Found ext -> ext | Not_found -> (!minus_infty, !minus_infty)
 ;;



(* first start after *)
let binary_search_tau index_data token k =
  if k = !plus_infty then (!plus_infty, !plus_infty) else
  try
    let positions = Hashtbl.find index_data token in
    let size = Array.length positions in
    if size = 0 then (!plus_infty, !plus_infty)
    else
      begin
      let first = ref 0 in
      let last = ref (size - 1)  in
      let mid = ref ((!first + !last) / 2) in
      while positions.(!mid) <> k && first <= last do
	if positions.(!mid) > k
	then last := !mid - 1 else first := !mid + 1 ;
	mid := (!first + !last) / 2
      done ;
      (* Recherche termine mais pas forcment russie. *)
      if positions.(!mid) = k then
	begin
        (* Commence exactement  k *)
	let p = positions.(!mid) in
        (* Gestion des infinis locaux *)
	if p > !minus_infty && p < !plus_infty then
	  let q = p + (Token.length token - 1) in
	  (p, q)
	else (!plus_infty, !plus_infty)
	end
      else
        (* On a forcement last = first - 1 et t.(!last) < k < t.(!first) *)
        (* Si le tableau ne contient pas d'lments < k alors !last est  *)
        (* en dehors des limites du tableau (et vaut donc -1).           *)
        (* C.f The Art of Computer Programming Vol 3 (D. Knuth)          *)
        (* Donc on renvoie l'lment commenant > e s'il n'est pas en    *)
        (* dehors du tableau auquel cas il faut retourner (infty, infty) *)
	if !first >= size then (!plus_infty, !plus_infty)
	else
	  begin
	  let p = positions.(!first) in
	  (* Gestion des infinis locaux *)
	  if p > !minus_infty && p < !plus_infty then
	    let q = p + (Token.length token - 1) in
	    (p, q)
	  else (!plus_infty, !plus_infty)
	  end
      end
  with Not_found -> (!plus_infty, !plus_infty)
;;



(* first end after *)
(* Remarque: rho s k = tau s (k - (strlen(s) -1)) *)
let binary_search_rho index_data token k =
  if k = !plus_infty then (!plus_infty, !plus_infty) else
  let str_len = (Token.length token - 1) in
  let k' = (if k = !minus_infty then k
            else if k >= str_len then (k - str_len)
                 else !minus_infty) in
  binary_search_tau index_data token k'
;;




(* last starting before *)
let binary_search_rho' index_data token k =
  if k = !minus_infty then (!minus_infty, !minus_infty) else
  try
    let positions = Hashtbl.find index_data token in
    let size = Array.length positions in
    if size = 0 then (!minus_infty, !minus_infty)
    else
      begin
      let first = ref 0 in
      let last = ref (size - 1)  in
      let mid = ref ((!first + !last) / 2) in
      while positions.(!mid) <> k && first <= last do
	if positions.(!mid) > k
	then last := !mid - 1 else first := !mid + 1 ;
	mid := (!first + !last) / 2
      done ;
      (* Recherche termine mais pas forcment russie. *)
      if positions.(!mid) = k then
	begin
        (* Commence exactement  k *)
	let p = positions.(!mid) in
        (* Gestion des infinis locaux *)
	if p > !minus_infty && p < !plus_infty then
	  let q = p + (Token.length token - 1) in
	  (p, q)
	else (!minus_infty, !minus_infty)
	end
      else
        (* On a forcement last = first - 1 et t.(!last) < k < t.(!first) *)
        (* Si le tableau ne contient pas d'lments < k alors !last est  *)
        (* en dehors des limites du tableau (et vaut donc -1).           *)
        (* C.f The Art of Computer Programming Vol 3 (D. Knuth)          *)
        (* Donc on renvoie l'lment commenant > e s'il n'est pas en    *)
        (* dehors du tableau auquel cas il faut retourner (infty, infty) *)
	if !last < 0 then (!minus_infty, !minus_infty)
	else
	  begin
	  let p = positions.(!last) in
	  (* Gestion des infinis locaux *)
	  if p > !minus_infty && p < !plus_infty then
	    let q = p + (Token.length token - 1) in
	    (p, q)
	  else (!minus_infty, !minus_infty)
	  end
      end
  with Not_found -> (!minus_infty, !minus_infty)
;;



(* last ending before *)
(* Remarque: tau' s k = rho' s (k - (strlen(s) -1)) *)
let binary_search_tau' index_data token k =
  if k = !minus_infty then (!minus_infty, !minus_infty) else
  let str_len = (Token.length token - 1) in
  (* Ne pas oublier que la soustraction est modulo l'infini *)
  let k' = (if k = !plus_infty then k
            else if k >= str_len then (k - str_len)
                 else !minus_infty) in
  binary_search_rho' index_data token k'
;;









(* Version de debug pour comparer les rsultats de la recherche *)
(* squentielle avec ceux de la recherche dichotomique.         *)
let tau index_data token k =
  let r1 = sequential_search_tau index_data token k in
  let r2 = binary_search_tau index_data token k in
  if r1 <> r2 then
    begin
    Printf.eprintf "BUG tau at k = (%d, %s) Expected (%d,%d) Got (%d,%d)\n"
                  k (Token.string token) (fst r1) (snd r1) (fst r2) (snd r2) ;
    flush stderr ;
    assert false
    end ;
  r1
;;
let rho index_data token k =
  let r1 = sequential_search_rho index_data token k in
  let r2 = binary_search_rho index_data token k in
  if r1 <> r2 then
    begin
    Printf.eprintf "BUG rho(%d, %s) Expected (%d,%d) Got (%d,%d)\n"
                  k (Token.string token) (fst r1) (snd r1) (fst r2) (snd r2) ;
    flush stderr ;
    assert false
    end ;
  r1
;;


let rho' index_data token k =
  let r1 = sequential_search_rho' index_data token k in
  let r2 = binary_search_rho' index_data token k in
  if r1 <> r2 then
    begin
    Printf.eprintf "BUG rho'(%d, %s) Expected (%d,%d) Got (%d,%d)\n"
                  k (Token.string token) (fst r1) (snd r1) (fst r2) (snd r2) ;
    flush stderr ;
    assert false
    end ;
  r1
;;
let tau' index_data token k =
  let r1 = sequential_search_tau' index_data token k in
  let r2 = binary_search_tau' index_data token k in
  if r1 <> r2 then
    begin
    Printf.eprintf "BUG tau'(%d, %s) Expected (%d,%d) Got (%d,%d)\n"
                  k (Token.string token) (fst r1) (snd r1) (fst r2) (snd r2) ;
    flush stderr ;
    assert false
    end ;
  r1
;;
