(********************************************************************** *)
(*                                                                      *)
(*                        Ubi Est - Indexing package                    *)
(*                                                                      *)
(*  Copyright Franois Pessaux 1999-2007.                               *)
(*  Distributed according to the LICENCE file contained in the archive. *)
(*                                                                      *)
(************************************************************************)


open Infinity ;;


(* Arbre de syntaxe des requtes *)
type request =
  | Simple of Token.t
  | And of request * request
  | Or of request * request
  | Followed_by of request * request
  | Contained_in of request * request
  | Containing of request * request
  | Not_contained_in of request * request
  | Not_containing of request * request
;; 


(* **************************
val plus_epsilon : int -> int

Incrmentation modulo +/- l'infini.
************************** *)
let plus_epsilon x =
  if x = !plus_infty then !plus_infty
  else if x = !minus_infty then !minus_infty
  else x + 1
;;



(* **************************
val minus_epsilon : int -> int

Dcrmentation modulo +/- l'infini.
************************** *)
let minus_epsilon x =
  if x = !plus_infty then !plus_infty
  else if x <= !minus_infty then !minus_infty
  else x - 1
;;



(* **************************
val tau :
  Index.index_data -> request -> Gclist.extent_position -> Gclist.extent
************************** *)
let rec tau index_data request k =
  if k = !plus_infty then (!plus_infty, !plus_infty)
  else
    match request with
     | Simple token -> Gclist.tau index_data token k
     | Followed_by (a, b) ->
	 let (p, q) = tau index_data a k in
	 if p = !plus_infty then (!plus_infty, !plus_infty)
	 else
	   let (p', q') = tau index_data b (plus_epsilon q) in
	   if p' = !plus_infty then (!plus_infty, !plus_infty)
	   else
	     let (p'', q'') = tau' index_data a (minus_epsilon p') in
	     (p'', q')
     | And (a, b) ->
	 let (p, q) = tau index_data a k in
	 if p = !plus_infty then (!plus_infty, !plus_infty)
	 else
	   let (p', q') = tau index_data b k in
	   if p' = !plus_infty then (!plus_infty, !plus_infty)
	   else
	     let (p'', q'') = tau' index_data a (max q q') in
	     let (p''', q''') = tau' index_data b (max q q') in
	     ((min p'' p'''), (max q'' q'''))
     | Or (a, b) ->
	 let (p, q) = tau index_data a k in
	 let (p', q') = tau index_data b k in
	 if q < q' then (p, q)
	 else if q > q' then (p', q')
	 else ((max p p'), q)
     | Contained_in (a, b) ->
	 let (p, q) = tau index_data a k in
	 let (p', q') = rho index_data b q in
	 if p' <= p then (p, q) else tau index_data request p'
     | Containing (a, b) ->
	 let (p, q) = tau index_data a k in
	 rho index_data request q
     | Not_contained_in (a, b) ->
	 let (p, q) = tau index_data a k in
	 let (p', q') = rho index_data b q in
	 if p' > p then (p, q) else rho index_data request (plus_epsilon q')
     | Not_containing (a, b) ->
	 let (p, q) = tau index_data a k in
	 rho index_data request q



(* **************************
val tau' :
  Index.index_data -> request -> Gclist.extent_position -> Gclist.extent
************************** *)
and tau' index_data request k =
  if k = !minus_infty then (!minus_infty, !minus_infty)
  else
    match request with
     | Simple token -> Gclist.tau' index_data token k
     | Followed_by (a, b) ->
	 let (p, q) = tau' index_data b k in
	 if p = !minus_infty then (!minus_infty, !minus_infty)
	 else
           let (p', q') = tau' index_data a (minus_epsilon p) in
	   if p' = !minus_infty then (!minus_infty, !minus_infty)
	   else
	     let (p'', q'') = tau index_data b (plus_epsilon q') in
	     (p', q'')
     | And (a, b) ->
	 let (p, q) = tau' index_data a k in
	 if p = !minus_infty then (!minus_infty, !minus_infty)
	 else
	   let (p', q') = tau' index_data b k in
	   if p' = !minus_infty then (!minus_infty, !minus_infty)
	   else
	     let (p'', q'') = tau index_data a (min p p') in
	     let (p''', q''') = tau index_data b (min p p') in
	     ((min p'' p'''), (max q'' q'''))
     | Or (a, b) ->
	 let (p, q) = tau' index_data a k in
	 let (p', q') = tau' index_data b k in
	 if p > p' then (p, q)
	 else if p < p' then (p', q')
	 else (p, min q q')
     | Contained_in (a, b) ->
	 let (p, q) = tau' index_data a k in
	 let (p', q') = rho' index_data b p in
	 if q' >= q then (p, q) else tau' index_data request q'
     | Containing (a, b) ->
	 let (p, q) = tau' index_data a k in
	 rho' index_data request p
     | Not_contained_in (a, b) ->
	 let (p, q) = tau' index_data a k in
	 let (p', q') = rho' index_data b p in
	 if q' < q then (p, q) else rho' index_data request (minus_epsilon p')
     | Not_containing (a, b) ->
	 let (p, q) = tau' index_data a k in
	 rho' index_data request p



(* **************************
val rho :
  Index.index_data -> request -> Gclist.extent_position -> Gclist.extent
************************** *)
and rho index_data request k =
  if k = !plus_infty then (!plus_infty, !plus_infty)
  else
    match request with
     | Simple token -> Gclist.rho index_data token k
     | Followed_by (a, b) ->
	 let (p, q) = tau' index_data request (minus_epsilon k) in
	 tau index_data (Followed_by (a, b)) (plus_epsilon p)
     | And (a, b) ->
	 let (p, q) = tau' index_data request (minus_epsilon k) in
	 tau index_data request (plus_epsilon p)
     | Or (a, b) ->
	 let (p, q) = rho index_data a k in
	 let (p', q') = rho index_data b k in
	 if q < q' then (p, q)
	 else if q > q' then (p', q')
	 else ((max p p'), q)
     | Contained_in (a, b) ->
	 let (p, q) = rho index_data a k in
	 tau index_data request p
     | Containing (a, b) ->
	 let (p, q) = rho index_data a k in
	 let (p', q') = tau index_data b p in
	 if q' <= q then (p, q) else rho index_data request q'
     | Not_contained_in (a, b) ->
	 let (p, q) = rho index_data a k in
	 tau index_data request p
     | Not_containing (a, b) ->
	 let (p, q) = rho index_data a k in
	 let (p', q') = tau index_data b p in
	 if q' > q then (p, q) else tau index_data request (plus_epsilon p')


(* **************************
val rho' :
  Index.index_data -> request -> Gclist.extent_position -> Gclist.extent
************************** *)
and rho' index_data request k =
  if k = !minus_infty then (!minus_infty, !minus_infty)
  else
    match request with
     | Simple token -> Gclist.rho' index_data token k
     | And (a, b) ->
	 let (p, q) = tau index_data request (plus_epsilon k) in
	 tau' index_data request (minus_epsilon q)
     | Or (a, b) ->
	 let (p, q) = tau index_data request (plus_epsilon k) in
	 tau' index_data request (minus_epsilon q)
     | Followed_by (a, b) ->
	 let (p, q) = tau index_data request (plus_epsilon k) in
	 tau' index_data request (minus_epsilon q)
     | Contained_in (a, b) ->
	 let (p, q) = rho' index_data a k in
	 tau' index_data request q
     | Containing (a, b) ->
	 let (p, q) = rho' index_data a k in
	 let (p', q') = tau' index_data b q in
	 if p' >= p then (p, q) else rho' index_data request p'
     | Not_contained_in (a, b) ->
	 let (p, q) = rho' index_data a k in
	 tau' index_data request q
     | Not_containing (a, b) ->
	 let (p, q) = rho' index_data a k in
	 let (p', q') = tau' index_data b q in
	 if p' < p then (p, q) else tau' index_data request (minus_epsilon q')
;;



(* **************************
val verify_request : Stoplist.t -> request -> bool

Vrifie que la requte ne fait pas intervenir de mots ignors par la stoplist.
Auquel cas rien ne sera trouv, ou le rsultat sera fauss. Retourne 'true'
si la requte est valide, sinon 'false'.
************************** *)
let rec verify_request stoplist = function
  | Simple (Token.Text txt) -> not (Stoplist.mem txt stoplist)
  | Simple _ -> true
  | And (a, b) ->
      verify_request stoplist a && verify_request stoplist b
  | Or (a, b) ->
      verify_request stoplist a && verify_request stoplist b
  | Followed_by (a, b) ->
      verify_request stoplist a && verify_request stoplist b
  | Contained_in (a, b) ->
      verify_request stoplist a && verify_request stoplist b
  | Containing (a, b) ->
      verify_request stoplist a && verify_request stoplist b
  | Not_contained_in (a, b) ->
      verify_request stoplist a && verify_request stoplist b
  | Not_containing (a, b) ->
      verify_request stoplist a && verify_request stoplist b
;;



(* **************************
val driver :
  request ->
  Index.t -> (Gclist.extent_position * Gclist.extent_position) list

Mainloop de recherche des occurences d'une requte dans toute la base de
donnes.
************************** *)
let driver req index =
  let index_data = index.Index.data in
  let res_gc_list = ref [] in
  let current_pos = ref (0) in
    (* On fait une requte par fichier et restreinte aux limites de *)
    (* ce fichier.                                                  *)
  for i = 0 to (Array.length index.Index.map) - 1 do
    (* Rcupre l'intervalle occup par le fichier courant *)
    let ((local_minf, local_pinf), _) = index.Index.map.(i) in
    minus_infty := local_minf - 1 ;
    plus_infty := local_pinf + 1 ;
    current_pos := local_minf ;
    (* On peut maintenant traiter cette requte *)
    let go_on = ref true in
    while !go_on do
      let (p, q) = tau index_data req !current_pos in
      if p = !plus_infty then go_on := false
      else
	begin
        (* Attention, la liste des rsultats est construite en ordre *)
        (* inverse.                                                  *)
	res_gc_list := (p,q) :: !res_gc_list ;
	current_pos := (plus_epsilon p)
	end
    done ;
  done ;
  List.rev !res_gc_list
;;

