(********************************************************************** *)
(*                                                                      *)
(*                        Ubi Est - Indexing package                    *)
(*                                                                      *)
(*  Copyright Franois Pessaux 1999-2007.                               *)
(*  Distributed according to the LICENCE file contained in the archive. *)
(*                                                                      *)
(************************************************************************)


type temporary_index =
    (Token.t, Index.file_position list ref) Hashtbl.t ;;


(* Position de dpart dans les fichiers virtuellement concatns *)
let global_file_base = ref 0 ;;
let tmp_index_data = Hashtbl.create 13 ;;
let recursive_indexing = ref false ;;
let guess_indexing = ref false ;;
let forget_unknown = ref false ;;
let tmp_mapping = ref [] ;;                (* Doit tre inverse avant usage *)
exception Skip_file of string ;;
let ignored_regexprs = ref [] ;;
let ignored_files_regexprs = ref ([] :Str.regexp list) ;;
(* Liste des fichiers de la ligne de commande *)
let cmdline_filenames = ref ([] : string list) ;;
(* Liste des fichiers comportant des listes de fichiers  indexer *)
let files_filenames = ref ([] : string list) ;;
(* Liste des fichiers de stoplist (liste   *)
(* de mots  ignorer durant l'indexation). *)
let stoplist_filenames = ref ([] : string list) ;;
(* Nom du fichier de sortie par dfaut de l'index. *)
let output_filename = ref "index.out" ;;



type preprocess =
  | No_pp
  | External_pp of string
;;



let prepro_table =
  [ (Fileformat.Unknown_format, No_pp) ;
    (Fileformat.Postscript, (External_pp "ps2ascii *")) ;
    (Fileformat.Pdf, (External_pp "pdftotext *")) ;
    (Fileformat.Lha, (External_pp "lha l *")) ;
    (Fileformat.Dvi, No_pp) ;
    (Fileformat.Latex, No_pp) ;
    (Fileformat.MS_Office_doc, (External_pp "antiword *")) ;
    (Fileformat.Rar, No_pp) ;
    (Fileformat.Zip, (External_pp "unzip -t *")) ;
    (Fileformat.Gzip, (External_pp "gzip -l *")) ;
    (Fileformat.Compress, No_pp) ;
    (Fileformat.Arj, (External_pp "unarj l *")) ;
    (Fileformat.Msdos_exe, No_pp) ;
    (Fileformat.Gif, No_pp) ;
    (Fileformat.Png, No_pp) ;
    (Fileformat.Jpeg, No_pp) ;
    (Fileformat.Bmp, No_pp) ;
    (Fileformat.Xpm, No_pp) ;
    (Fileformat.Riff_wav, No_pp) ;
    (Fileformat.Mod, No_pp) ;
    (Fileformat.Realaudio, No_pp) ;
    (Fileformat.Iff, No_pp) ;
    (Fileformat.Fig, No_pp) ;
    (Fileformat.Phigs, No_pp) ;
    (Fileformat.Cgm, No_pp) ;
    (Fileformat.Sit, No_pp) ;
    (Fileformat.Rtf, (External_pp "antiword *")) ;
    (Fileformat.Html, No_pp) ;
    (Fileformat.Tar, (External_pp "tar tvzf *")) ;
    (Fileformat.Java_bytecode, No_pp) ;
    (Fileformat.C_H, No_pp) ;
    (Fileformat.PBM_image_text, No_pp) ;
    (Fileformat.PGM_image_text, No_pp) ;
    (Fileformat.PPM_image_text, No_pp) ;
    (Fileformat.PBM_image_rawbits, No_pp) ;
    (Fileformat.PGM_image_rawbits, No_pp) ;
    (Fileformat.PPM_image_rawbits, No_pp) ;
    (Fileformat.PAM_image, No_pp) ;
    (Fileformat.ELF_32_bits_exe, No_pp) ;
    (Fileformat.ELF_64_bits_exe, No_pp) ;
    (Fileformat.OCAML_Cmi, No_pp) ;
    (Fileformat.OCAML_Cmo, No_pp) ;
    (Fileformat.OCAML_Cmx, No_pp) ;
    (Fileformat.OCAML_Cmi, No_pp) ;
    (Fileformat.OCAML_Cma, No_pp) ;
    (Fileformat.OCAML_Cmxa, No_pp)
  ] ;;



(* **************************************************************
   [Fonc] find_lexer : string -> Lexing.lexbuf -> Token.t * int
   [Description] : Dtermination du lexer d'indexation  utiliser
     en fonction du format du fichier. Gre en dur les fichiers
     'tar' (octet 257 : "ustar").
   ************************************************************** *)
let find_lexer_and_pp filename =
  if not !guess_indexing then (Lexplaintext.file, None)
  else
    let file_type = Fformat.get filename in
    let file_type_message = Fileformat.to_string file_type in
    Printf.printf "File type looks like %s\n" file_type_message ;
    let lexer =
      match file_type with
       | Fileformat.Html -> Lexhtml.file
       | Fileformat.C_H -> Lexcfile.file
       | Fileformat.Latex -> Lexlatex.file
       | Fileformat.MS_Office_doc -> Lexplaintext.file
       | Fileformat.Arj | Fileformat.Msdos_exe | Fileformat.Riff_wav
       | Fileformat.Mod | Fileformat.Realaudio | Fileformat.Tar
       | Fileformat.Postscript | Fileformat.Pdf | Fileformat.Lha
       | Fileformat.Dvi | Fileformat.Rar | Fileformat.Zip
       | Fileformat.Gzip | Fileformat.Compress | Fileformat.Java_bytecode
       | Fileformat.Gif | Fileformat.Png | Fileformat.Jpeg | Fileformat.Bmp
       | Fileformat.Xpm | Fileformat.Iff | Fileformat.Fig | Fileformat.Phigs
       | Fileformat.Cgm | Fileformat.Sit | Fileformat.Rtf
       | Fileformat.PBM_image_text | Fileformat.PGM_image_text
       | Fileformat.PPM_image_text | Fileformat.PBM_image_rawbits
       | Fileformat.PGM_image_rawbits | Fileformat.PPM_image_rawbits
       | Fileformat.PAM_image | Fileformat.ELF_32_bits_exe
       | Fileformat.ELF_64_bits_exe | Fileformat.OCAML_Cmi
       | Fileformat.OCAML_Cmo | Fileformat.OCAML_Cmx | Fileformat.OCAML_Cma
       | Fileformat.OCAML_Cmxa -> raise (Skip_file file_type_message)
       | Fileformat.Unknown_format ->
	   (* On vrifie si on doit indexer les fichiers dont *)
	   (* on n'a pas russi a dterminer le type.         *)
	   if !forget_unknown then raise (Skip_file "Unknown")
	   else Lexplaintext.file in
    let preprocessed_filename =
      match List.assoc file_type prepro_table with
       | _ -> None
	     (* Preprocessign pour plus tard. *)
    (* | No_pp -> None
       | External_pp pp ->
	   begin
	   let tmpfile = Filename.temp_file "ubiestpp" "" in
	   let prefix_pp = Index.build_command_call pp filename 0 in
	   let cmd = prefix_pp ^ " > " ^ tmpfile in
	   Printf.printf "Preprocessing as '%s'\n" cmd ;
	   if Sys.command cmd <> 0 then
	     begin
             (try Sys.remove tmpfile with Sys_error _ -> ()) ;
             raise (Skip_file (Fileformat.to_string file_type)) ;
	     end ;
	   Some tmpfile
	   end *) in
    (lexer, preprocessed_filename)
;;



(* **********************************************************************
   [Fonc] create_index_from_file :
             (Token.t, int list ref) Hashtbl.t -> string -> string * int
   [Description] : Ajout des mots du fichier 'filename' dans la structure
        temporaire d'index.
   ********************************************************************** *)
let create_index_from_file tmp_index_data filename =
  (* Dtermination du format du fichier. *)
  let (lexer_entry_point, preprocessed_filename) =
    find_lexer_and_pp filename in
  let file =
    open_in
      (match preprocessed_filename with None -> filename 
       | Some fname -> fname) in
  let lexbuf = Lexing.from_channel file in
  (* C'est parti... *)
  let continue = ref true in
  let global_ending_position = ref 0 in
  while !continue do
    (* Mot lex et sa position locale au fichier courant. *)
    let (word, local_position) = lexer_entry_point lexbuf in
    (* Position globale dans la base de donne vue comme 1 seul fichier.  *)
    let global_position = !global_file_base + local_position in
    match word with
     | Token.Markup Markup.Eof ->
	 (* Le caractre EOF doit tre ignor.        *)
	 (* Ainsi ici 'position' = taille du fichier. *)
	 global_ending_position := global_position - 1 ;
	 continue := false
     | _ ->
	 try
	   let entry = Hashtbl.find tmp_index_data word in
           (* Attention : les positions sont stockes en ordre inverse. *)
	   entry := global_position :: !entry
	 with
	 | Not_found -> Hashtbl.add tmp_index_data word (ref [global_position])
  done ;
  close_in file ;
  global_file_base := !global_ending_position + 1 ;
  (* Si l'indexation a t faite dans un fichier     *)
  (* pr-process, on dtruit le fichier temporaire. *)
  (match preprocessed_filename with None -> ()
   | Some fname -> try Sys.remove fname with Sys_error _ -> ()) ;
  (* Retourne le nom du fichier d'origine (NON pr-process) *)
  (* et sa position de fin au niveau global.                 *)
  (filename, !global_file_base - 1)
;;



(* *****************************************************************
   [Fonc] build_position_mapping :
       (string * int) list -> ((int * int) * string) array
   [Description] : Construction de la table permettant de mapper des
      positions dans la base de donnes sur des positions dans les
      fichiers contenus dans cette base de donnes.
   ***************************************************************** *)
let build_position_mapping ends =
  let global_start_pos = ref 0 in
  let size = List.length ends in
  let i = ref 0 in
  let position_mapping = Array.create size ((-1, -1), "") in
  List.iter
    (fun (filename, global_end_pos) ->
      position_mapping.(!i) <-
	((!global_start_pos, global_end_pos), filename) ;
      global_start_pos := global_end_pos + 1 ;
      incr i)
    ends ;
  position_mapping
;;



(* *********************************************************************
   [Fonc] build_index_data :
             ('a, 'b list ref) Hashtbl.t -> ('a, 'b array) Hashtbl.t
   [Description] : Construction de la structure de donnes d'index
       finale  partir de la structure temporaire. Ca revient en fait
        transformer les listes de positions des occurrences en
       tableaux de positions (avec inversion au passage).
   ********************************************************************* *)
let build_index_data tmp_data =
  let data = Hashtbl.create 13 in
  Hashtbl.iter
    (fun token positions ->
      (* Ne pas oublier de renverser la liste des occurrences. *)
      Hashtbl.add data token (Array.of_list (List.rev !positions)))
    tmp_data ;
  data
;;



(* ************************************************************************
   [Fonc] treat_file : Index.file_name -> unit
   [Description] : Traitement d'un fichier  indexer. Gestion de la
        descente rcursive dans les rpertoires. L'ajout dans la structure
        temporaire d'index est fait par effet de bord.
   ************************************************************************* *)
let rec treat_file filename =
  (* On regarde si le nom du fichier correspond  une des regexps de *)
  (* noms de fichiers  ignorer.  noter que cette vrification est  *)
  (* prioritaire sur la liste des noms de fichiers  indexer s'il y  *)
  (* en a une de fournie.                                            *)
  let to_be_ignored =
    let bn = Filename.basename filename in
    List.exists (fun r -> Str.string_match r bn 0) !ignored_files_regexprs in
  if not to_be_ignored then
    begin
    (* On ne doit pas ignorer ce fichier... *)
    Printf.printf "Dealing with file '%s'\n" filename ;
    try
      let file_stat = Unix.stat filename in
      match file_stat.Unix.st_kind with
       | Unix.S_LNK ->
	   Printf.printf "Skipping file '%s' (symbolic link).\n"  filename
       | Unix.S_DIR ->
	   if not !recursive_indexing then
	     Printf.printf "Skipping directory.\n"
	   else
	     begin
	     Printf.printf "Descending in directory.\n" ;
	     let dir_handle = Unix.opendir filename in
	     (try
	       while true do
		 let filename' = Unix.readdir dir_handle in
		 if filename' <> "." && filename' <> ".." then
		   treat_file (filename ^ "/" ^ filename')
	       done
	     with End_of_file -> Unix.closedir dir_handle)
	     end
       | Unix.S_REG ->
	   begin
	   try
	     if file_stat.Unix.st_size > 0 then
	       begin
	       let map_info = create_index_from_file tmp_index_data filename in
               (* Attention on mmorise en sens inverse. *)
	       tmp_mapping := map_info :: !tmp_mapping
	       end
	     else Printf.printf "Skipped empty file '%s'.\n" filename ;
	   with Skip_file fmt -> Printf.printf "Skipped '%s'.\n" fmt
	   end
       | _ -> Printf.printf "Skipping file '%s' (non regular file).\n" filename
    with
    | Unix.Unix_error (_, m1, m2) ->
	Printf.eprintf "Unix_error '%s' : '%s' on '%s'\n" m1 m2 filename
    | Sys_error m ->
	Printf.eprintf "Sys_error : '%s' on '%s'\n" m filename ;
    end
;;



(* ****************************************************************
   [Fonc] treat_file_list : string -> unit
   [Description] : Indexation de tous les fichiers dont le nom sont
       contenus dans le fichier de nom 'list_filename'.
   **************************************************************** *)
let treat_file_list list_filename =
  Printf.printf "Getting files list from '%s'.\n" list_filename ;
  let list_file = open_in list_filename in
  let go_on = ref true in
  while !go_on do
    try
      let filename = input_line list_file in
      (* Les commentaires commencent par  # en dbut de ligne. *)
      if filename <> "" && filename.[0] <> '#' then treat_file filename
    with End_of_file -> go_on := false
  done ;
  close_in list_file
;;



(* *************************************************************
   [Fonc] load_stop_list : string -> unit
   [Description] : Chargement d'une stop list  utiliser lors de
       l'indexation.
   ************************************************************* *)
let load_stop_list filename =
  let file = open_in filename in
  try
    while true do
      let word = input_line file in
      if word <> "" then
	GlobalStoplist.stop_list := Stoplist.add word !GlobalStoplist.stop_list
    done
  with End_of_file ->
    Printf.printf "Stop-list loaded from '%s'.\n" filename ;
    close_in file
;;



let go () =
  (* En premier, on charge les stoplists. *)
  List.iter load_stop_list !stoplist_filenames ;
  (* On construit la liste de regexprs  ignorer lors de l'indexation. *)
  ignored_files_regexprs := List.map Str.regexp !ignored_regexprs ;
  (* Maintenant, on indexe rellement les fichiers de donnes. *)
  List.iter (fun fname -> treat_file fname) (List.rev !cmdline_filenames) ;
  (* Maintenant au tour des fichiers provenant de -F *)
  List.iter treat_file_list (List.rev !files_filenames) ;
  (* Construction de la structure d'index finale. *)
  if !tmp_mapping <> [] then
    begin
    (* Il y a effectivement eu des fichiers indexs. *)
    let final_index = {
      Index.stoplist = !GlobalStoplist.stop_list ;
      Index.map =
        build_position_mapping (List.rev !tmp_mapping) ;
      Index.data = build_index_data tmp_index_data } in
    (* Maintenant, sauve le rsultat de l'indexation sur disque. *)
    (try
      let output_file = open_out_bin !output_filename in
      output_value output_file (final_index : Index.t) ;
      (* Penser  clore le fichier sinon il risque de rester vide. *)
      close_out output_file
    with Sys_error m -> 
      Printf.eprintf "Can't write output file : %s\n" m) ;
    Printf.printf "Index creation is succesfully finished.\n"
    end
  else Printf.eprintf "createindex refuses to create an empty index.\n"
;;
