(********************************************************************** *)
(*                                                                      *)
(*                        Ubi Est - Indexing package                    *)
(*                                                                      *)
(*  Copyright Franois Pessaux 1999-2007.                               *)
(*  Distributed according to the LICENCE file contained in the archive. *)
(*                                                                      *)
(************************************************************************)


(* *********************************************************************
   [Exception] : No_index_loaded
   [Description] : Sert  notifier le fait qu'aucun index n'a t charg
      en mmoire et qu'il est donc impossible de lancer une requte.
   [Rem] : Non export hors de ce module.
           Normalement cette exception ne doit pas sortir en dehors des
           fonctions de ce module.
   ********************************************************************* *)
exception No_index_loaded ;;



(* *************************************************************************
   [Type] selection
   [Description] : Comme on ne peut pas rcuprer la slection courante dans
       une clist on mmorise ici toutes les infos ncessaires pour retrouver
       les caractristiques du 'hit' slectionn dans la liste des 'hits'.
   [Champs] :
     - row : Numro de la ligne slectionne (commence  0).
     - filename : Nom du fichier dans lequel le 'hit' se trouve.
     - exact_start : Dbut de l'extend matchant exactement la requte.
     - exact_stop : Fin de l'extend matchant exactement la requte.
     - actual_start : Dbut de l'affichage rel (avec extention). Ce champ
          est mutable car l'utilisateur peut demander  tendre le contexte
          de citation.
     - actual_stop : Fin de l'affichage rel (avec extention). Ce champe est
          mutable pour la mme raison que celui ci-dessus.
   [Rem] : non export hors de ce module.
   ************************************************************************* *)
type selection = {
  row : int ;
  filename : string ;
  exact_start : int ;
  exact_stop : int ;
  mutable actual_start : int ;
  mutable actual_stop : int
 } ;;



(* *********************************************************************
   [Valeur] current_selection : (selection option) ref
   [Description] : Variable globale permettant de mmoriser l'ventuelle
       slection de 'hit' courante.
   [Rem] : Non exporte hors de ce module.
   ********************************************************************* *)
let current_selection = ref (None : selection option) ;;



class all_callbacks = object (self)
  inherit Gui_glade_callbacks.default_callbacks

   (* ***************************************************************
      [Fonc] on_information_menu_activate : unit -> unit
      [Fonc] on_project_menu_activate : unit -> unit
      [Description] : Dsactive le callback par dfaut de MlGlade qui
         affiche un message de dbug lors d'un click sur un menu.
      [Args] :
        - unit
      [Retour] :
        - unit
      *************************************************************** *)
  method on_information_menu_activate () = ()
  method on_project_menu_activate () = ()



   (* ***************************************************************
      [Fonc] on_quit_menu_item_activate : unit -> unit
      [Description] : Callback invoqu lors du clic sur l'item de
          menu de sortie du programme.
      [Args] :
        - unit
      [Retour] :
        - unit
      *************************************************************** *)
  method on_quit_menu_item_activate () = GtkMain.Main.quit ()



   (* ***************************************************************
      [Fonc] on_load_index_menu_item_activate : unit -> unit
      [Description] : Callback invoqu lors du clic sur l'item de
          chargement d'un fichier d'index. Il ouvre un requester de
          fichier et essaye de charger le fichier d'index en cas de
          validation de l'utilisateur. En cas d'erreur, une fentre
          de message est affiche. Si tout se passe bien, la variable
          globale servant  rfrencer l'index est mise  jour par
          effet de bord.
      [Args] :
        - unit
      [Retour] :
        - unit
      *************************************************************** *)
  method on_load_index_menu_item_activate () =
    begin
    let file_selector =
      GWindow.file_selection
	~title: "Open index" ~position:`CENTER
	~modal: true ~filename: !Guiglobal.current_dir () in 
    (* Callback pour la validation du chargement. *)
    let ok_callback () =
      let filename = file_selector#get_filename in
      (* Mmorisation du repertoire courant. *)
      Guiglobal.current_dir := (Filename.dirname filename) ^ "/" ;
      if filename <> "" then
	(try
	  let input_file = open_in_bin filename in
          (* Mise  jour de la rfrence sur l'index par effet de bord. *)
	  Guiglobal.global_index :=
	    Some ((input_value input_file) : Index.t) ;
	  close_in input_file ;
	  (* Mise  jour de la barre de status. *)
	  let statusbar_context =
	    (match !Guiglobal.main_statusbar_context with Some c -> c
	     | None -> assert false) in
	  ignore (statusbar_context#push ("Current index is : " ^ filename))
	with
	| Sys_error errmsg -> Convenient.error_notify self errmsg
	| Failure ("input_value: bad object") ->
	    Convenient.error_notify self "Invalid index file format.") ;
      file_selector#destroy () in
    (* On affiche effectivement la fentre de filerequester. *)
    file_selector#show () ;
    ignore
      (file_selector#cancel_button#connect#clicked
	 ~callback: (file_selector#destroy)) ;
    ignore (file_selector#ok_button#connect#clicked  ~callback: ok_callback)
    end



  method on_menu_item_about_activate () =
    begin
    let about_w = new Gui_glade_interface.top_about_window self in
    self#set_about_window about_w ;
    about_w#about_version_label#set_text Index.version_release ;
    ignore
      (GtkBase.Widget.add_events about_w#about_window#as_widget
	 [`ALL_EVENTS]) ;
    about_w#about_window#show()
    end



  method on_create_index_menu_item_activate () =
    begin
    (* On commence par r-initialiser les variables globales de Create. *)
    Create.recursive_indexing := false ;
    Create.guess_indexing := false ;
    Create.forget_unknown := false ;
    Create.ignored_regexprs := [] ;
    Create.cmdline_filenames := [] ;
    Create.files_filenames := [] ;
    Create.stoplist_filenames := [] ;
    Create.output_filename := "index.out" ;
    (* Maintenant, on attaque les ajouts de Clists  l'interface. *)
    let create_w = new Gui_glade_interface.top_create_window self in
    self#set_create_window create_w ;
    let (files_clist : string GList.clist) =
      GList.clist
	~packing: (create_w#scrolledwindow6#add_with_viewport)
	~titles_show: true ~titles: ["File name"] () in
    files_clist#set_column ~min_width: 200 ~resizeable: true 0 ;
    files_clist#set_selection_mode `MULTIPLE ;
    Guiglobal.files_clist := Some files_clist ;
    let (files_files_clist : string GList.clist) =
      GList.clist
	~packing: (create_w#scrolledwindow8#add_with_viewport)
	~titles_show: true ~titles: ["File name"] () in
    files_files_clist#set_column ~min_width: 200 ~resizeable: true 0 ;
    files_files_clist#set_selection_mode `MULTIPLE ;
    Guiglobal.files_files_clist := Some files_files_clist ;
    let (stoplists_clist : string GList.clist) =
      GList.clist
	~packing: (create_w#scrolledwindow9#add_with_viewport)
	~titles_show: true ~titles: ["File name"] () in
    stoplists_clist#set_column ~min_width: 200 ~resizeable: true 0 ;
    stoplists_clist#set_selection_mode `MULTIPLE ;
    Guiglobal.stoplists_clist := Some stoplists_clist ;
    create_w#index_output_filename_entry#set_text "index.out" ;
    ignore
      (GtkBase.Widget.add_events
	 create_w#create_window#as_widget [`ALL_EVENTS]) ;
    create_w#create_window#show()
    end



  method on_menu_item_help_activate () =
    begin
    try
      ignore
	(Sys.command
	   "netscape /usr/share/ubiest/doc/UbiEst_Documentation.html &")
    with Sys_error err -> Convenient.error_notify self err
    end



  method on_about_close_button_clicked () =
    self#top_about_window#about_window#destroy ()


  method on_cancel_create_button_clicked () =
    self#top_create_window#create_window#destroy ()



  method on_window_error_close_button_clicked () =
    self#top_window_error#window_error#destroy ()



  method on_window_warning_close_button_clicked () =
    self#top_window_warning#window_warning#destroy ()



  method on_window_info_close_button_clicked () =
    self#top_window_info#window_info#destroy ()



  (* ******************************************************************
     [Fonc] on_submit_button_clicked : unit -> unit
     [Description] : Callback invoqu lros du clic sur le bouton de
         soumission d'une requte. Le texte de la requte est rcupr
         dans la widget 'text', puis pars et calcul. En cas d'erreur,
         un message est affich.
     [Args] :
       - unit
     [Retour] :
       - unit
     ****************************************************************** *)
  method on_submit_button_clicked () =
    try
      (* On commence par vider les anciens rsultats. *)
      let result_clist =
	(match !Guiglobal.result_clist with
	 | None -> assert false
	 | Some w -> w) in
      result_clist#clear () ;
      current_selection := None ;
      (* Il faut penser a rajouter un '\n' a la fin sinon le parser *)
      (* se plaint car la fin de requte est justement un '\n'.     *)
      let text_widget = self#top_main_window#request_text in
      let request_string =
	(text_widget#get_chars ~start: 0 ~stop: text_widget#length) ^ "\n" in
      let lexbuf = Lexing.from_string request_string in
      let request = Parserequest.request Lexrequest.request lexbuf in
      let index = (match !Guiglobal.global_index with Some i -> i
       | _ -> raise No_index_loaded) in
      if Request.verify_request index.Index.stoplist request then
	begin
	(* Et, hop on se met  calculer. *)
	let gclist = Request.driver request index in
	(* Affichage des rsultats. *)
	let context_widget = self#top_main_window#context_text in
	(* Commence par vider la fentre de contexte. *)
	context_widget#delete_text ~start: 0 ~stop: (context_widget#length) ;
	(*  ne pas oublier avant chaque nouvelle requte, *)
	(* sinon on pte les invariants du file-mapping !  *)
	Filemap.map_extend_init () ;
	(* Maintenant on peut appeler 'map_extend_on_file' pour toutes *)
	(* les occurrences de la requte courante.                     *)
	let nb_hits = ref 0 in
	List.iter
	  (fun extent ->
            let ((fstart, fstop), filename) as position_mapping =
	      Filemap.map_extend_on_file index extent in
	    (* Ajoute le hit dans la liste. *)
	    ignore
	      (result_clist#append
		 [filename; string_of_int fstart; string_of_int fstop]) ;
	    (* Ajoute une rfrence vers la location de cette entre. *)
	    result_clist#set_row_data !nb_hits position_mapping ;
	    incr nb_hits)
	  gclist ;
	(* Affiche le nombre de hits dansle titre de la frame de rsultats *)
	self#top_main_window#results_frame#set_label
	  ("Query results (" ^ (string_of_int !nb_hits) ^ " hits)")
	end
      else
	begin
	let window_warning = new Gui_glade_interface.top_window_warning self in
	self#set_window_warning window_warning ;
	window_warning#warning_label#set_text
	  "Aborted request containing words ignored during indexing." ;
	ignore
	  (GtkBase.Widget.add_events 
	     window_warning#window_warning#as_widget [`ALL_EVENTS]) ;
	window_warning#window_warning#show ()
	end
    with exn ->
      let msg =
	(match exn with
	 | Lexrequest.Lexical_error -> "Lexical error in submitted query."
	 | Parsing.Parse_error -> "Syntax error in submitted query."
	 | No_index_loaded -> "No index loaded."
	 | e -> raise e) in
      let window_error = new Gui_glade_interface.top_window_error self in
      self#set_window_error window_error ;
      window_error#error_label#set_text msg ;
      ignore
	(GtkBase.Widget.add_events 
	   window_error#window_error#as_widget [`ALL_EVENTS]) ;
      window_error#window_error#show ()



  method click_on_row_result_binding
      ~row: row ~column: column ~event: (_ : GdkEvent.Button.t option) =
    (* Si column est -1, alors c'est que c'est la slection automatique *)
    (* du systme qui a appel le callback. C'est du au fait qu'on est  *)
    (* en mode de slection BROWSE afin d'viter que cliquer 2 fois sur *)
    (* une ligne ne la d-slectionne.                                  *)
    if column <> -1 then
      begin
      (* Commence par vider la fentre de contexte. *)
      let context_widget = self#top_main_window#context_text in
      context_widget#delete_text ~start: 0 ~stop: (context_widget#length) ;
      (* On rcupre les infos de la ligne slectionne. *)
      let result_clist =
	(match !Guiglobal.result_clist with
	 | None -> assert false
	 | Some w -> w) in
      let ((fstart, fstop), filename) = result_clist#get_row_data row in
      try
	let file = open_in filename in
	seek_in file fstart ;
	let quote_size = (fstop - fstart) + 1 in
	let quote = String.create quote_size in
	ignore (input file quote 0 quote_size) ;
	close_in file ;
	context_widget#insert quote ;
	(* Mise  jour de la slection courante. *)
	current_selection :=
	  Some { row = row ; filename = filename ;
		 exact_start = fstart ; exact_stop = fstop ;
		 actual_start = fstart ; actual_stop = fstop }
      with Sys_error msg ->
	current_selection := None ;
	let window_error = new Gui_glade_interface.top_window_error self in
	self#set_window_error window_error ;
	window_error#error_label#set_text
	  ("Can't quote because:\n\"" ^ msg ^ "\"\n") ;
	ignore
	  (GtkBase.Widget.add_events 
	     window_error#window_error#as_widget [`ALL_EVENTS]) ;
	window_error#window_error#show ()
      end



  method on_send_to_button_clicked () =
    match !current_selection with
     | Some sel ->
	 begin
	 try
	   let file = open_in sel.filename in
	   let fformat_lexbuf = Lexing.from_channel file in
	   let format = Fileformat.file fformat_lexbuf in
	   close_in file ;
	   (* Hard coded *)
	   let command =
	     (try
	       List.assoc format !Preferences.global_prefs.Preferences.viewers
	     with Not_found -> "") in
	   ignore (Sys.command
		     (Index.build_command_call
			command sel.filename sel.exact_start))
	 with Sys_error err -> Convenient.error_notify self err
	 end
     | None ->
	 let window_error = new Gui_glade_interface.top_window_error self in
	 self#set_window_error window_error ;
	 window_error#error_label#set_text "No selected hit" ;
	 ignore
	   (GtkBase.Widget.add_events 
	      window_error#window_error#as_widget [`ALL_EVENTS]) ;
	 window_error#window_error#show ()



  method on_file_info_button_clicked () =
    match !current_selection with
     | Some sel ->
	 begin
	 try
	   let format = Fileformat.to_string (Fformat.get sel.filename) in
	   let file = open_in sel.filename in
	   let size = string_of_int (in_channel_length file) in
	   close_in file ;
	   Convenient.info_notify self
	     ("File : '" ^ sel.filename ^ "'\nSize: " ^ size ^ "\nType: "
	      ^ format)
	 with Sys_error err -> Convenient.error_notify self err
	 end
     |  None -> Convenient.error_notify self "No selected document"



  method on_increase_context_button_clicked () =
    begin
    match !current_selection with
     | Some sel ->
	 begin
	 try
	   let file = open_in sel.filename in
	   (* Nouvelle tendue du contexte. *)
	   let (new_actual_start, new_actual_stop) =
	     ((max 0 (sel.actual_start - 80)),
	      (min (sel.actual_stop + 80) (in_channel_length file))) in
	   (* Primo effacer l'ancien contexte. *)
	   let context_widget = self#top_main_window#context_text in
	   context_widget#delete_text ~start: 0 ~stop: (context_widget#length) ;
	   (* Extension au dbut *)
	   seek_in file new_actual_start ;
	   let size_prefix = sel.exact_start - new_actual_start in
	   let context_prefix = String.create size_prefix in
	   ignore (input file context_prefix 0 size_prefix) ;
	   (* Contexte rel, c'est--dire du motif exact. *)
	   let size_exact = 1 + sel.exact_stop - sel.exact_start in
	   let context_exact = String.create size_exact in
	   ignore (input file context_exact 0 size_exact) ;
	   (* Extention  la fin. *)
	   let size_suffix = new_actual_stop - (sel.exact_stop + 1) in
	   let context_suffix = String.create size_suffix in
	   ignore (input file context_suffix 0 size_suffix) ;
	   (* Termin ... *)
	   close_in file ;
	   (match !Preferences.global_prefs.Preferences.context_font with
	    | None -> context_widget#insert context_prefix
	    | Some f -> context_widget#insert ~font: f context_prefix) ;
	   (match !Preferences.global_prefs.Preferences.exact_font with
	    | None -> context_widget#insert context_exact
	    | Some f -> context_widget#insert ~font: f context_exact) ;
	   (match !Preferences.global_prefs.Preferences.context_font with
	    | None -> context_widget#insert context_suffix
	    | Some f -> context_widget#insert ~font: f context_suffix) ;
	   (* On met  jour l'tendue du contexte avec les nouvelles valeurs. *)
	   sel.actual_start <- new_actual_start ;
	   sel.actual_stop <- new_actual_stop
	 with Sys_error err -> Convenient.error_notify self err
	 end
     |  None -> Convenient.error_notify self "No selected document"
    end



  method on_save_context_button_clicked () =
    begin
    let context_widget = self#top_main_window#context_text in
    if context_widget#length = 0 then
      Convenient.error_notify self "Context is empty : nothing to save."
    else
      let file_selector =
	GWindow.file_selection
	  ~title: "Save context to file" ~position:`CENTER
	  ~modal: true ~filename: !Guiglobal.current_dir () in 
      (* On affiche effectivement la fentre... *)
      file_selector#show () ;
      ignore
	(file_selector#cancel_button#connect#clicked
	   ~callback: (file_selector#destroy)) ;
      let ok_callback () =
	let filename = file_selector # get_filename in
	(* Mmorisation du repertoire courant. *)
	Guiglobal.current_dir := (Filename.dirname filename) ^"/" ;
	if filename <> "" then
	  (try
	    let output_file = open_out filename in
	    let msg =
	      context_widget#get_chars
		~start: 0 ~stop: (context_widget#length) in
	    output_string output_file msg ;
	    close_out output_file
	  with Sys_error err -> Convenient.error_notify self err) ;
	file_selector#destroy () in
      ignore (file_selector#ok_button#connect#clicked  ~callback: ok_callback)
    end



  method click_on_row_viewer_select_binding selection viewers_clist
      ~row: row ~column: column ~event: (_ : GdkEvent.Button.t option) =
    let viewer_entry = self#top_preferences_window#application_name_entry in
    viewer_entry#set_editable true ;
    let app_name = snd (viewers_clist#get_row_data row) in
    (* On modifie par effet de bord le numro le la ligne *)
    (* slctionne pour le que callback se trouvant dans *)
    (* 'on_preferences_menu_item_activate' puisse savoir  *)
    (* a posteriori qui est slctionn.                  *)
    selection := Some (row) ;
    viewer_entry#set_text app_name



  method on_cancel_prefs_button_clicked () =
    self#top_preferences_window#preferences_window#destroy ()



  method on_apply_prefs_button_clicked
      (viewers_clist : (Fileformat.format * string) GList.clist) () =
    begin
    (* On rcupre le nom temporaire des fontes. "Temporaire", car *)
    (* il se peut que l'on n'arrive pas  les ouvrir. Dans ce cas, *)
    (* le vritable nom sera celui remis  "" lors du traitement   *)
    (* de l'erreur de chargement de chaque fonte.                  *)
    let tmp_exact_font_name =
      self#top_preferences_window#exact_context_font_name_label#text in
    let tmp_context_font_name =
      self#top_preferences_window#extended_context_font_name_label#text in
    let (exact_font, exact_font_name) =
      (try ((Some (Gdk.Font.load_fontset tmp_exact_font_name)),
	    tmp_exact_font_name) with
	Gpointer.Null ->
	  Convenient.error_notify
	    self ("Unsupported font '" ^ tmp_exact_font_name ^
		  "' for exact context highlight.") ;
	  (None, "")) in
    let (context_font, context_font_name) =
      (try ((Some (Gdk.Font.load_fontset tmp_context_font_name)),
	    tmp_context_font_name) with
	Gpointer.Null ->
	  Convenient.error_notify
	    self ("Unsupported font '" ^ tmp_context_font_name ^
		  "' for extended context highlight.") ;
	  (None, "")) in
    let nb_rows = viewers_clist#rows in
    let rec get n =
      if n = nb_rows then []
      else (viewers_clist#get_row_data n) :: (get (n + 1)) in
    let viewers = get 0 in
    let new_prefs = {
      Preferences.exact_font_name = exact_font_name ;
      Preferences.context_font_name = context_font_name ;
      Preferences.exact_font = exact_font ;
      Preferences.context_font = context_font ;
      Preferences.viewers = viewers } in
    Preferences.global_prefs := new_prefs
    end



  method on_save_prefs_button_clicked
      (viewers_clist : (Fileformat.format * string) GList.clist) () =
    self#on_apply_prefs_button_clicked viewers_clist () ;
    (try
      Preferences.save_prefs () ;
      Convenient.info_notify self "Preferences saved\n"
    with
    | Preferences.No_home_dir ->
	Convenient.error_notify
	  self
	  "Can't save preferences : can't find user home directory."
    | Sys_error err ->
	Convenient.error_notify self ("Can't save preferences. " ^ err)
    | other ->
	Convenient.error_notify
	  self ("Can't save preferences. " ^ (Printexc.to_string other) ^
		" .")) ;
    self#top_preferences_window#preferences_window#destroy ()    



  method on_preferences_menu_item_activate () =
    let prefs_w = new Gui_glade_interface.top_preferences_window self in
    self#set_preferences_window prefs_w ;
    (* Mise  jour initiale des labels de noms de fontes. *)
    prefs_w#exact_context_font_name_label#set_text
      !Preferences.global_prefs.Preferences.exact_font_name ;
    prefs_w#extended_context_font_name_label#set_text
      !Preferences.global_prefs.Preferences.context_font_name ;
    (* Rajout  la main de la clist de resultat car non gre par MlGlade. *)
    let (viewers_clist : (Fileformat.format * string) GList.clist) =
      GList.clist_poly
	~packing: (prefs_w#scrolledwindow7#add_with_viewport)
	~titles_show: true ~titles: ["File type" ; "Viewer" ; "(N/A) Preprocessor"]
	() in
    viewers_clist#set_column ~min_width: 150 ~resizeable: true 0 ;
    viewers_clist#set_column ~min_width: 300 ~resizeable: true 1 ;
    viewers_clist#set_column ~min_width: 300 ~resizeable: true 2 ;
    viewers_clist#set_selection_mode `SINGLE ;
    (* Mise  jour initiale de la liste des viewers. *)
    let viewers_cpt = ref 0 in
    List.iter
      (fun (file_type, viewer) ->
	ignore
	  (viewers_clist#append [Fileformat.to_string file_type; viewer]) ;
	viewers_clist#set_row_data !viewers_cpt (file_type, viewer) ;
        incr viewers_cpt)
      !Preferences.global_prefs.Preferences.viewers ;
    (* Connexion du clic de slection sur le liste des viewers. *)
    let selection = ref None in
    ignore
      (viewers_clist#connect#select_row
	 ~callback: (self#click_on_row_viewer_select_binding
		       selection viewers_clist)) ;
    (* Callback de validation de l'entry de viewer. *)
    let set_viewer () =
      begin
      let viewer_entry = prefs_w#application_name_entry in
      match !selection with
       | None -> ()
       | Some row  ->
	   let new_command = viewer_entry#text in
	   if String.contains new_command '*' then
	     begin
	     viewers_clist#set_cell ~text: new_command row 1 ;
	     let file_type = fst (viewers_clist#get_row_data row) in
	     viewers_clist#set_row_data row (file_type, new_command)
	     end
	   else
	     Convenient.error_notify self
	       "Can't find '*' (filename specification)\n in command line."
      end in
    ignore
      (prefs_w#application_name_entry#connect#activate
	 ~callback: set_viewer) ;
    (* Connexion du callback du bouton d'application des prfrences. *)
    ignore
      (prefs_w#apply_prefs_button#connect#clicked
	 ~callback: (self#on_apply_prefs_button_clicked viewers_clist)) ;
    (* Connexion du callback du bouton de sauvegarde des prfrences. *)
    ignore
      (prefs_w#save_prefs_button#connect#clicked
	 ~callback: (self#on_save_prefs_button_clicked viewers_clist)) ;
    (* Affichage final de la fentre de prfrences. *)
    ignore
      (GtkBase.Widget.add_events prefs_w#preferences_window#as_widget
	 [`ALL_EVENTS]) ;
    prefs_w#preferences_window#show()



  method on_font_sel_ok_button_clicked related_label () =
    begin
    let font_selection_dialog =
      self#top_font_selection_dialog#font_selection_dialog in
    let font_name = font_selection_dialog#selection#font_name in
    self#top_font_selection_dialog#font_selection_dialog#destroy () ;
    match font_name with
     | None -> ()
     | Some fn_name -> related_label#set_text fn_name
    end



  method factorized_on_font_button_click related_label =
    let fontsel_w = new Gui_glade_interface.top_font_selection_dialog self in
    self#set_font_selection_dialog fontsel_w ;
    ignore
      (fontsel_w#font_sel_cancel_button#connect#clicked
	 ~callback: (self#on_font_sel_ok_button_clicked related_label)) ;
    ignore
      (GtkBase.Widget.add_events fontsel_w#font_selection_dialog#as_widget
	 [`ALL_EVENTS]) ;
    fontsel_w#font_selection_dialog#show()
 


  method on_exact_context_sel_button_clicked () =
    self#factorized_on_font_button_click
      (self#top_preferences_window#exact_context_font_name_label)



  method on_extended_context_sel_button_clicked () =
    self#factorized_on_font_button_click
      (self#top_preferences_window#extended_context_font_name_label)



  method on_font_sel_cancel_button_clicked () =
    self#top_font_selection_dialog#font_selection_dialog#destroy ()



  method on_add_files_general_button_clicked related_title related_clist  =
    begin
    let file_selector =
      GWindow.file_selection
	~title: related_title ~position:`CENTER
	~modal: true ~filename: !Guiglobal.current_dir () in
    (* On tend la liste du file_selector pour avoir de la multi-slection. *)
    let file_list = file_selector#file_list in
    file_list#set_selection_mode `EXTENDED ;
    (* Callback pour la validation du chargement. *)
    let ok_callback () =
      let filename = file_selector#get_filename in
      (* Mmorisation du repertoire courant. *)
      let current_rep = (Filename.dirname filename) ^"/" in
      Guiglobal.current_dir := current_rep ;
      (* On initialise un flag temporaire servant  savoir si l'utilisateur *)
      (* a effectivement slectionn un ou des noms de fichiers dans la     *)
      (* liste du slecteur. Si ce n'est pas le cas et que '#get_filename'  *)
      (* ne rend pas "", alors on considrera que l'utilisateur veut        *)
      (* inclure rcursivement tous les fichiers  partir du rpertoire     *)
      (* dont le nom nous est renvoy par '#get_filename'.                  *)
      let just_add_files = ref false in
      (* On parcours chacune des lignes de la slection de   *)
      (* fichier, en ajoutant celles qui sont slctionnes. *)
      for i = 0 to (file_list#rows - 1) do
	match file_list#get_row_state i with
	 | `ACTIVE | `SELECTED ->
	     begin
	     (* On a bien des fichiers slectionns, donc on ne considre *)
	     (* pas que l'utilisateur veut inclure rcursivement  partir *)
	     (* du rpertoire dsign par ce que rend '#get_filename'.    *)
	     just_add_files := true ;
	     let fname = current_rep ^ (file_list#cell_text i 0) in
	     ignore (related_clist#append [fname])
	     end
	 | `NORMAL -> ()
	 | `PRELIGHT -> ()
	 | `INSENSITIVE -> ()
      done ;
      (* On traite finalement le cas o l'utilisateur a voulu inclure *)
      (* des fichiers rcursivement  partir d'un rpertoire.         *)
      if filename <> "" && (not !just_add_files) then
	ignore (related_clist#append [filename]) ;
      file_selector#destroy () in
    (* On affiche effectivement la fentre de filerequester. *)
    file_selector#show () ;
    ignore
      (file_selector#cancel_button#connect#clicked
	 ~callback: (file_selector#destroy)) ;
    ignore (file_selector#ok_button#connect#clicked  ~callback: ok_callback)
    end



  method on_remove_files_general_button_clicked related_clist =
    (* Attention: il faut retirer les entres de la liste en partant *)
    (* de la dernire car si oncommence par la premire, une fois    *)
    (* celle-ci supprime, le nom dntres de la liste est diminu   *)
    (* de 1, et le numro de "row" qui tait 1 avant est devenu 0.   *)
    (* autrement procder en ordre croissant induit un dcalage     *)
    (* chaque suppression, et ce n'est pas ce que l'on veut !!!      *)
    for i = (related_clist#rows - 1) downto 0  do
      match related_clist#get_row_state i with
       | `ACTIVE | `SELECTED ->related_clist#remove ~row:i
       | _ -> ()
    done
 


  method on_remove_files_button_clicked () =
    let files_clist =
      (match !Guiglobal.files_clist with Some cl -> cl
       | None -> assert false) in
    self#on_remove_files_general_button_clicked files_clist



  method on_remove_files_files_clicked () =
    let files_files_clist =
      (match !Guiglobal.files_files_clist with Some cl -> cl
       | None -> assert false) in
    self#on_remove_files_general_button_clicked files_files_clist



  method on_remove_stoplist_files_button_clicked () =
    let stoplists_clist =
      (match !Guiglobal.stoplists_clist with Some cl -> cl
       | None -> assert false) in
    self#on_remove_files_general_button_clicked stoplists_clist



  method on_add_files_button_clicked () =
    let files_clist =
      (match !Guiglobal.files_clist with Some cl -> cl
       | None -> assert false) in
    self#on_add_files_general_button_clicked "Add file (s)" files_clist



  method on_add_files_files_button_clicked () =
    let files_files_clist =
      (match !Guiglobal.files_files_clist with Some cl -> cl
       | None -> assert false) in
    self#on_add_files_general_button_clicked
      "Add file (s) from file(s)" files_files_clist



  method on_add_stoplist_files_button_clicked () =
    let stoplists_clist =
      (match !Guiglobal.stoplists_clist with Some cl -> cl
       | None -> assert false) in
    self#on_add_files_general_button_clicked
      "Add stoplist file (s)" stoplists_clist



  method on_force_unknown_checkbutton_toggled () =
    Create.forget_unknown := not !Create.forget_unknown



  method on_guess_checkbutton_toggled () =
    Create.guess_indexing := not !Create.guess_indexing



  method on_recursive_checkbutton_toggled () =
    Create.recursive_indexing := not !Create.recursive_indexing



  method on_proceed_create_button_clicked () =
    begin
    (* On rcupre le nom de sortie de l'index. *)
    Create.output_filename :=
      self#top_create_window#index_output_filename_entry#text ;
    (* On rcupre la liste des fichiers  indexer. *)
    let files_clist =
      (match !Guiglobal.files_clist with Some cl -> cl
       | None -> assert false) in
    for i = 0 to (files_clist#rows - 1) do
      Create.cmdline_filenames :=
	(files_clist#cell_text i 0) :: !Create.cmdline_filenames
    done ;
    (* On rcupre la liste des fichiers dcrivant les fichiers  indexer. *)
    let files_files_clist =
      (match !Guiglobal.files_files_clist with Some cl -> cl
       | None -> assert false) in
    for i = 0 to (files_files_clist#rows - 1) do
      Create.files_filenames :=
	(files_files_clist#cell_text i 0) :: !Create.files_filenames
    done ;
    (* On rcupre la liste des fichiers de stoplist. *)
    let stoplists_clist =
      (match !Guiglobal.stoplists_clist with Some cl -> cl
       | None -> assert false) in
    for i = 0 to (stoplists_clist#rows - 1) do
      Create.stoplist_filenames :=
	(stoplists_clist#cell_text i 0) :: !Create.stoplist_filenames
    done ;
    (* On rcupre les expressions rgulires  ignorer. *)
    let accu = ref "" in
    let regexps_text_widget = self#top_create_window#ignored_regexps_text in
    String.iter
      (function
	| '\n' -> if !accu <> "" then
	    begin
	    Create.ignored_regexprs := !accu :: !Create.ignored_regexprs ;
	    accu := ""
	    end
	| whatever -> accu := !accu ^ (String.make 1 whatever))
      (regexps_text_widget#get_chars
	 ~start: 0 ~stop: regexps_text_widget#length) ;
    (* On rcupre l'ventuelle dernire chane si   *)
    (* elle n'a pas t close par un retour-chariot. *)
    if !accu <> "" then
      Create.ignored_regexprs := !accu :: !Create.ignored_regexprs ;
    (* Et on lance la cration de l'index. *)
    Create.go () ;
    self#top_create_window#create_window#destroy ()
    end



  method on_index_out_file_sel_button_clicked () =
    begin
    let file_selector =
      GWindow.file_selection
	~title: "Save future index under" ~position:`CENTER
	~modal: true ~filename: !Guiglobal.current_dir () in 
    (* Callback pour la validation du chargement. *)
    let ok_callback () =
      let filename = file_selector#get_filename in
      (* Mmorisation du repertoire courant. *)
      Guiglobal.current_dir := (Filename.dirname filename) ^"/" ;
      if filename <> "" then
	self#top_create_window#index_output_filename_entry#set_text filename ;
      file_selector#destroy () in
    (* On affiche effectivement la fentre de filerequester. *)
    file_selector#show () ;
    ignore
      (file_selector#cancel_button#connect#clicked
	 ~callback: (file_selector#destroy)) ;
    ignore (file_selector#ok_button#connect#clicked  ~callback: ok_callback)
    end
	
end
