(********************************************************************** *)
(*                                                                      *)
(*                        Ubi Est - Indexing package                    *)
(*                                                                      *)
(*  Copyright Franois Pessaux 1999-2007.                               *)
(*  Distributed according to the LICENCE file contained in the archive. *)
(*                                                                      *)
(************************************************************************)


(* ***********************************************************************
   [Type] t
   [Description] : Structure permettant de mmoriser les prfrences de
       l'utilisateur.
   [Champs] :
     - exact_font_name : Nom de fonte de la partie exacte du contexte
          de citation.
     - exact_font : Fonte Gdk de la partie exacte du contexte de citation.
     - context_font_name : Nom de fonte utilise pour les extensions de
          contexte de citation.
     - context_font : Fonte utilise pour les extensions de contexte de
          citation.
     - viewers : Table des viewers. C'est une liste qui associe un type
          de fichier avec le format de commande  invoquer pour visualiser
          ce fichier. Dans ce format, le caractre '*' sera remplac par
          le nom effectif du fichier  visualiser, avant le lancement de
          la commande.
   [Rem] : Export en clair hors de ce module.
   *********************************************************************** *)
type t = {
  exact_font_name : string ;
  exact_font : Gdk.font option ;
  context_font_name : string ;
  context_font : Gdk.font option ;
  viewers : (Fileformat.format * string) list
  }
;;



(* *************************************************************************
   [Valeur] global_prefs : t ref
   [Description] : Variable globale o sont stockes les prefrences actives.
   [Rem] : Exporte hors de ce module.
   ************************************************************************* *)
let global_prefs =
  ref {
  exact_font_name = "-adobe-*-*-*-*-*-12-*-*-*-*-*-*-*" ;
  exact_font = None ;
  context_font_name = "fixed" ;
  context_font = None ;
  viewers = [ (Fileformat.Unknown_format, "emacs * --eval=\"(goto-char (+ 1 @))\"") ;
	      (Fileformat.Postscript, "gv *") ;
	      (Fileformat.Pdf, "acroread *") ;
	      (Fileformat.Lha, "lha l *") ;
	      (Fileformat.Dvi, "xdvi *") ;
	      (Fileformat.Latex, "emacs * --eval=\"(goto-char (+ 1 @))\"") ;
	      (Fileformat.MS_Office_doc, "soffice *") ;
	      (Fileformat.Rar, "") ;
	      (Fileformat.Zip, "unzip -t") ;
	      (Fileformat.Gzip, "gzip -l *") ;
	      (Fileformat.Compress, "") ;
	      (Fileformat.Arj, "unarj l *") ;
	      (Fileformat.Msdos_exe, "") ;
	      (Fileformat.Gif, "ee *") ;
	      (Fileformat.Png, "ee *") ;
	      (Fileformat.Jpeg, "ee *") ;
	      (Fileformat.Bmp, "ee *") ;
	      (Fileformat.Xpm, "ee *") ;
	      (Fileformat.Riff_wav, "cat * > /dev/dsp") ;
	      (Fileformat.Mod, "") ;
	      (Fileformat.Realaudio, "") ;
	      (Fileformat.Iff, "ee *") ;
	      (Fileformat.Fig, "xfig *") ;
	      (Fileformat.Phigs, "") ;
	      (Fileformat.Cgm, "") ;
	      (Fileformat.Sit, "") ;
	      (Fileformat.Rtf, "emacs * --eval=\"(goto-char (+ 1 @))\"") ;
 	      (Fileformat.Html, "netscape *") ;
	      (Fileformat.Tar, "tar tvzf *") ;
	      (Fileformat.Java_bytecode, "") ;
	      (Fileformat.C_H, "emacs * --eval=\"(goto-char (+ 1 @))\"") ;
	      (Fileformat.PBM_image_text, "ee *") ;
	      (Fileformat.PGM_image_text, "ee *") ;
	      (Fileformat.PPM_image_text, "ee *") ;
	      (Fileformat.PBM_image_rawbits, "ee *") ;
	      (Fileformat.PGM_image_rawbits, "ee *") ;
	      (Fileformat.PPM_image_rawbits, "ee *") ;
	      (Fileformat.PAM_image, "ee *") ;
	      (Fileformat.ELF_32_bits_exe, "") ;
	      (Fileformat.ELF_64_bits_exe, "") ;
	      (Fileformat.OCAML_Cmi, "") ;
	      (Fileformat.OCAML_Cmo, "") ;
	      (Fileformat.OCAML_Cmx, "") ;
	      (Fileformat.OCAML_Cmi, "") ;
	      (Fileformat.OCAML_Cma, "") ;
	      (Fileformat.OCAML_Cmxa, "")
	    ]
 }
;;



(* *********************************************************************
   [Exception] No_home_dir
   [Description] : Sert  signaler qu'il est impossible de dterminer le
        homedir de l'utilisateur.
   [Rem] : Exporte hors de ce module.
   ********************************************************************* *)
exception No_home_dir ;;



(* *********************************************************************
   [Fonc] : gethomedir : unit -> string
   [Description] : Fonction charge de retrouver le home directory de
       l'utilisateur. On commence d'abord par regarder la variable
       d'environnement '$HOME'. Si on ne le trouve pas, on va rechercher
       l'uid de l'utilisateur et on va parser le fichier '/etc/passwd'
       pour retrouver le champ homedir correspondant.
   [Args] :
     - Nant
   [Retour] :
     - string : La chane de caractres reprsentant le chemin complet
         du homedir de l'utilisateur.
   [Rem] : Exporte hors de ce module.
           Peut lever l'exception No_home_dir si l'on n'a pas russi 
           dterminer le homedir de l'utilisateur, ni  partir de la
           variable d'environnement, ni  partir du '/etc/passwd'.
   ********************************************************************* *)
let gethomedir () =
  try Sys.getenv "HOME" with
  | Not_found ->
      begin
      let uid = Unix.getuid () in
      let etc_passwd = try open_in "/etc/passwd" with _ -> raise No_home_dir in
      let home_dir = ref "" in
      try
	while !home_dir = "" do
	  let line = input_line etc_passwd in
	  (* Saute au password. *)
	  let i = String.index line ':' in
	  (* Saute au dbut de l'uid. *)
	  let i' = String.index_from line (i + 1) ':' in
	  (* Saute au dbut du gid *)
	  let i'' = String.index_from line (i' + 1) ':' in
	  let id_string = String.sub line (i' + 1) (i'' - i' - 1) in
	  let id = int_of_string id_string in
	  if id = uid then
	    begin
            (* Saute  la fin du gid. *)
	    let i3 = String.index_from line (i'' + 1) ':' in
            (* Saute au dbut du homedir. *)
	    let i4 = String.index_from line (i3 + 1) ':' in
            (* Fin du homedir. *)
	    let i5 = String.index_from line (i4 + 1) ':' in
	    home_dir := String.sub line (i4 + 1) (i5 - i4 - 1) ;
	    end
	done ;
	close_in etc_passwd ;
	!home_dir
      with _ -> raise No_home_dir
      end
;;



(* *********************************************************************
   [Fonc]  save_prefs : unit -> unit
   [Description] : Fonction permettant de sauver les prfrences
        courantes dans le fichier de configuration de l'utilisateur
        actuel. Ce fichier se trouve dans le rpertoire racine de
        l'utilisateur et se nomme '.mlindexrc'. Notons que cette
        fonction ne sauvegarde pas TOUTE la structure de prfrences et
        en particulier les 2 structures de fontes Gdk car ce sont des
        objets et 'output_value' ne les accepte pas. Donc pour ces
        fontes, on ne sauvegarde que leur nom, et on les re-crera
        ultrieurement selon le besoin.
   [Args] :
     - unit
   [Retour] :
     - unit
   [Rem] : Exporte hors de ce module.
********************************************************************* *)
let save_prefs () =
  let filename = (gethomedir ()) ^ "/.mlindexrc" in
  let file = open_out_bin filename in
  output_value file ((!global_prefs.exact_font_name) : string) ;
  output_value file ((!global_prefs.context_font_name) : string) ;
  output_value file ((!global_prefs.viewers) :
		       (Fileformat.format * string) list) ;
  close_out file
;;



let load_prefs () =
  try
    let filename = (gethomedir ()) ^ "/.mlindexrc" in
    let file = open_in_bin filename in
    let exact_font_name = ((input_value file) : string) in
    let context_font_name = ((input_value file) : string) in
    let viewers = ((input_value file) : (Fileformat.format * string) list) in
    let new_prefs = {
      exact_font_name = exact_font_name ;
      context_font_name = context_font_name ;
      exact_font = None ;
      context_font = None ;
      viewers = viewers } in
    close_in file ;
    global_prefs := new_prefs
  with _ -> ()
;;
